from datetime import date
from time import time
from typing import Tuple, Union
from uuid import uuid4

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from requests import Session
from requests.adapters import HTTPAdapter, Retry

from .schema import (
    AggregateCarrierResponse,
    CarrierInvoiceDonateResponse,
    CarrierInvoicesDetailResponse,
    CarrierInvoicesHeaderResponse,
    InvoiceDetailResponse,
    InvoiceHeaderResponse,
    LotteryNumberResponse,
    LoveCodeResponse,
)
from .utils import (
    build_api_url,
    check_api_error,
    sign,
    validate_invoice_number,
    validate_invoice_random,
    validate_invoice_term,
)


class AppAPIClient(object):
    def __init__(
        self,
        app_id: str,
        api_key: str,
        uuid: Union[str, None] = None,
        ts_tolerance: int = 20,
        max_retries: int = 20,
        skip_validation: bool = False,
        timeout: Union[float, Tuple[float, float], Tuple[float, None]] = (3, 1),
    ):
        self.app_id = app_id
        self.api_key = api_key
        self.uuid = str(uuid) if uuid else str(uuid4())
        if ts_tolerance < 10 or ts_tolerance > 180:
            raise ValueError("ts_tolerance must be between 10 and 180")
        self.ts_tolerance = ts_tolerance
        if not isinstance(skip_validation, bool):
            raise ValueError("skip_validation must be a boolean")
        self.skip_validation = skip_validation
        self.serial = 1
        self.session = Session()
        self.session.headers.update(
            {"Content-Type": "application/x-www-form-urlencoded"}
        )
        self.session.mount(
            "https://",
            HTTPAdapter(
                max_retries=Retry(
                    total=max_retries,
                    backoff_factor=0.1,
                    allowed_methods=["POST"],
                    status_forcelist=[500, 502, 503, 504],
                    raise_on_status=False,
                )
            ),
        )
        self.timeout = timeout

    def get_lottery_numbers(
        self, invoice_term: str
    ) -> Union[LotteryNumberResponse, dict]:
        """查詢中獎發票號碼清單 v0.2"""
        URL = build_api_url("invapp")
        VERSION = 0.2
        if not validate_invoice_term(invoice_term):
            raise ValueError(f"Invalid invoice_term: {invoice_term}")
        data = {
            "version": VERSION,
            "action": "QryWinningList",
            "invTerm": invoice_term,
            "UUID": self.uuid,
            "appID": self.app_id,
        }
        results = check_api_error(
            self.session.post(URL, data=data, timeout=self.timeout)
        )
        if not self.skip_validation:
            results = LotteryNumberResponse.parse_obj(results)
        return results

    def get_invoice_header(
        self,
        barcode_type: Literal["QRCode", "Barcode"],
        invoice_number: str,
        invoice_date: date,
    ) -> Union[InvoiceHeaderResponse, dict]:
        """查詢發票表頭 v0.5"""
        URL = build_api_url("invapp")
        VERSION = 0.5
        if barcode_type not in ("QRCode", "Barcode"):
            raise ValueError("Type must be 'QRCode' or 'Barcode'")
        if not validate_invoice_number(invoice_number):
            raise ValueError(f"Invalid invoice number: {invoice_number}")
        data = {
            "version": VERSION,
            "type": barcode_type,
            "invNum": invoice_number,
            "action": "qryInvHeader",
            "generation": "V2",
            "invDate": invoice_date.strftime("%Y/%m/%d"),
            "UUID": self.uuid,
            "appID": self.app_id,
        }
        results = check_api_error(
            self.session.post(URL, data=data, timeout=self.timeout)
        )
        if not self.skip_validation:
            results = InvoiceHeaderResponse.parse_obj(results)
        return results

    def get_invoice_detail(
        self,
        barcode_type: Literal["QRCode", "Barcode"],
        invoice_number: str,
        invoice_date: date,
        invoice_random: str,
        invoice_term: Union[str, None] = None,
        invoice_encrypt: Union[str, None] = None,
        seller_id: Union[str, None] = None,
    ):
        """
        查詢發票明細 v0.6
        `invoice_random`: 錯誤將僅回傳發票表頭，正確將回傳完整明細

        `barcode_type`: "Barcode"
        `invoice_number`: 發票號碼
        `invoice_date`: 發票日期
        `invoice_random`: 發票隨機碼 (4位)
        `invoice_term`: 發票期別， 3 位民國年+ 2 位雙數月份，如：11006

        or

        `barcode_type`: "QRCode"
        `invoice_number`: 發票號碼
        `invoice_date`: 發票日期
        `invoice_random`: 發票隨機碼 (4位)
        `invoice_encrypt`: 發票檢驗碼 (左側QRCode中，24位)
        `seller_id`: 商家統編
        """
        URL = build_api_url("invapp")
        VERSION = 0.6
        if barcode_type == "QRCode":
            if not invoice_encrypt:
                raise ValueError(
                    "invoice_encrypt is required when barcode_type is QRCode"
                )
            if not isinstance(invoice_encrypt, str) or len(invoice_encrypt) != 24:
                raise ValueError("invoice_encrypt must be a string of 24 characters")
            if not seller_id:
                raise ValueError("seller_id is required when barcode_type is QRCode")
        elif barcode_type == "Barcode":
            if not invoice_term:
                raise ValueError(
                    "invoice_term is required when barcode_type is Barcode"
                )
            if not validate_invoice_term(invoice_term):
                raise ValueError(f"Invalid invoice_term: {invoice_term}")
        else:
            raise ValueError("barcode_type must be 'QRCode' or 'Barcode'")
        if not validate_invoice_number(invoice_number):
            raise ValueError(f"Invalid invoice number: {invoice_number}")
        if not validate_invoice_random(invoice_random):
            raise ValueError(f"Invalid invoice random: {invoice_random}")
        data = {
            "version": VERSION,
            "type": barcode_type,
            "invNum": invoice_number,
            "action": "qryInvDetail",
            "generation": "V2",
            "invTerm": invoice_term,
            "invDate": invoice_date.strftime("%Y/%m/%d"),
            "encrypt": invoice_encrypt,
            "sellerID": seller_id,
            "UUID": self.uuid,
            "randomNumber": invoice_random,
            "appID": self.app_id,
        }
        results = check_api_error(
            self.session.post(URL, data=data, timeout=self.timeout)
        )
        if not self.skip_validation:
            results = InvoiceDetailResponse.parse_obj(results)
        return results

    def get_love_code(self, query: str) -> dict:
        """捐贈碼查詢 v0.2"""
        URL = build_api_url("lovecode")
        VERSION = 0.2
        data = {
            "version": VERSION,
            "qKey": query,
            "action": "qryLoveCode",
            "UUID": self.uuid,
            "appID": self.app_id,
        }
        results = check_api_error(
            self.session.post(URL, data=data, timeout=self.timeout)
        )
        if not self.skip_validation:
            results = LoveCodeResponse.parse_obj(results)
        return results

    def get_carrier_invoices_header(
        self,
        card_type: str,
        card_number: str,
        start_date: date,
        end_date: date,
        card_encrypt: str,
        only_winning: bool = False,
    ) -> Union[dict, CarrierInvoicesHeaderResponse]:
        """載具發票表頭查詢 v0.5"""
        URL = build_api_url("invserv")
        VERSION = 0.5
        data = {
            "version": VERSION,
            "cardType": card_type,
            "cardNo": card_number,
            "expTimeStamp": "2147483647",
            "action": "carrierInvChk",
            "timeStamp": int(time() + self.ts_tolerance),
            "startDate": start_date.strftime("%Y/%m/%d"),
            "endDate": end_date.strftime("%Y/%m/%d"),
            "onlyWinningInv": "Y" if only_winning else "N",
            "uuid": self.uuid,
            "appID": self.app_id,
            "cardEncrypt": card_encrypt,
        }
        results = check_api_error(
            self.session.post(URL, data=data, timeout=self.timeout)
        )
        if not self.skip_validation:
            results = CarrierInvoicesHeaderResponse.parse_obj(results)
        return results

    def get_carrier_invoices_detail(
        self,
        card_type: str,
        card_number: str,
        invoice_number: str,
        invoice_date: date,
        card_encrypt: str,
        seller_name: Union[str, None] = None,
        amount: Union[int, None] = None,
    ) -> Union[dict, CarrierInvoicesDetailResponse]:
        """載具發票明細查詢 v0.5"""
        URL = build_api_url("invserv")
        VERSION = 0.5
        if not validate_invoice_number(invoice_number):
            raise ValueError(f"Invalid invoice number: {invoice_number}")
        data = {
            "version": VERSION,
            "cardType": card_type,
            "cardNo": card_number,
            "expTimeStamp": "2147483647",
            "action": "carrierInvDetail",
            "timeStamp": int(time() + self.ts_tolerance),
            "invNum": invoice_number,
            "invDate": invoice_date.strftime("%Y/%m/%d"),
            "uuid": self.uuid,
            "sellerName": seller_name,
            "amount": amount,
            "appID": self.app_id,
            "cardEncrypt": card_encrypt,
        }
        results = check_api_error(
            self.session.post(URL, data=data, timeout=self.timeout)
        )
        if not self.skip_validation:
            results = CarrierInvoicesDetailResponse.parse_obj(results)
        return results

    def carrier_donate_invoice(
        self,
        card_type: str,
        card_number: str,
        invoice_date: date,
        invoice_number: str,
        love_code: str,
        card_encrypt: str,
    ):
        """載具發票捐贈 v0.1"""
        URL = build_api_url("donate")
        VERSION = 0.1
        if not validate_invoice_number(invoice_number):
            raise ValueError(f"Invalid invoice number: {invoice_number}")
        data = {
            "version": VERSION,
            "serial": f"{self.serial:0>10}",
            "cardType": card_type,
            "cardNo": card_number,
            "expTimeStamp": "2147483647",
            "action": "carrierInvDnt",
            "timeStamp": int(time() + self.ts_tolerance),
            "invDate": invoice_date.strftime("%Y/%m/%d"),
            "invNum": invoice_number,
            "npoBan": love_code,
            "uuid": self.uuid,
            "appID": self.app_id,
            "cardEncrypt": card_encrypt,
        }
        signature = sign(data, self.api_key)
        data["signature"] = signature
        self.serial += 1
        results = check_api_error(
            self.session.post(URL, data=data, timeout=self.timeout)
        )
        if not self.skip_validation:
            results = CarrierInvoiceDonateResponse.parse_obj(results)
        return results

    def get_aggregate_carrier(
        self,
        card_type: str,
        card_number: str,
        card_encrypt: str,
    ) -> Union[dict, AggregateCarrierResponse]:
        """手機條碼歸戶載具查詢 v1.0"""
        URL = build_api_url("carrier")
        VERSION = 1.0
        data = {
            "version": VERSION,
            "serial": f"{self.serial:0>10}",
            "action": "qryCarrierAgg",
            "cardType": card_type,
            "cardNo": card_number,
            "cardEncrypt": card_encrypt,
            "appID": self.app_id,
            "timeStamp": int(time() + self.ts_tolerance),
            "uuid": self.uuid,
        }
        signature = sign(data, self.api_key)
        data["signature"] = signature
        self.serial += 1
        results = check_api_error(
            self.session.post(URL, data=data, timeout=self.timeout)
        )
        if not self.skip_validation:
            results = AggregateCarrierResponse.parse_obj(results)
        return results



import pytest

from tw_invoice.exception import APIError

TEST_CODE = 1
TEST_MESSAGE = "Error Message"


def test_api_error():
    """Test APIError"""
    with pytest.raises(APIError):
        raise APIError(TEST_CODE, TEST_MESSAGE)

    error = APIError(TEST_CODE, TEST_MESSAGE)
    assert error.code == TEST_CODE
    assert error.message == TEST_MESSAGE
    assert str(TEST_CODE) in str(error)
    assert TEST_MESSAGE in str(error)



import pytest
from requests.exceptions import HTTPError
from requests.models import Response

from tw_invoice.exception import APIError
from tw_invoice.utils import (
    check_api_error,
    validate_invoice_number,
    validate_invoice_random,
    validate_invoice_term,
    validate_phone_barcode,
)


@pytest.fixture
def client_error():
    response = Response()
    response.status_code = 404
    return response


@pytest.fixture
def server_error():
    response = Response()
    response.status_code = 500
    return response


@pytest.fixture
def api_error():
    response = Response()
    response.status_code = 200
    response.json = lambda: {"code": 998, "msg": "appID 不符合規定 (停權或尚未申請)"}
    return response


@pytest.fixture
def api_success():
    response = Response()
    response.status_code = 200
    response.json = lambda: {"code": "200", "msg": "OK"}
    return response


def test_check_api_error(client_error, server_error, api_error, api_success):
    # Test invalid type
    with pytest.raises(TypeError):
        check_api_error(None)

    # Test client error
    with pytest.raises(HTTPError):
        check_api_error(client_error)

    # Test server error
    with pytest.raises(HTTPError):
        check_api_error(server_error)

    # Test API error
    with pytest.raises(APIError):
        check_api_error(api_error)

    data = check_api_error(api_success)
    assert data["code"] == "200"


def test_validate_invoice_number():
    assert not validate_invoice_number(None)
    assert not validate_invoice_number("")
    assert not validate_invoice_number("5ab562e60d9ba2ee")
    assert not validate_invoice_number("AB")
    assert not validate_invoice_number("12345678")
    assert not validate_invoice_number("ab12345678")
    assert validate_invoice_number("AB12345678")


def test_validate_invoice_random():
    assert not validate_invoice_random(None)
    assert not validate_invoice_random("")
    assert not validate_invoice_random("abcd")
    assert not validate_invoice_random("56789")
    assert validate_invoice_random("0123")


def test_validate_invoice_term():
    assert not validate_invoice_term(None)
    assert not validate_invoice_term("")
    assert not validate_invoice_term("5ab562e60d9ba2ee")
    assert not validate_invoice_term("202206")
    assert not validate_invoice_term("11105")
    assert validate_invoice_term("11106")


def test_validate_phone_barcode():
    assert not validate_phone_barcode(None)
    assert not validate_phone_barcode("")
    assert not validate_phone_barcode("5ab562e60d9ba2ee")
    assert not validate_phone_barcode("AB12+-.")
    assert validate_phone_barcode("/AB12+-.")



class APIError(Exception):
    """Base class for all API errors."""

    def __init__(self, code: int, message: str, *args: object) -> None:
        super().__init__(*args)
        self.code = code
        self.message = message

    def __str__(self) -> str:
        return f"<{self.code}> {self.message}"



import hashlib
import hmac
import re
from base64 import b64encode
from urllib.parse import urlencode, urljoin

from requests.models import Response

from .exception import APIError


def build_api_url(id: str) -> str:
    BASE_URL = "https://api.einvoice.nat.gov.tw"
    PATHS = {
        "invapp": "/PB2CAPIVAN/invapp/InvApp",
        "lovecode": "/PB2CAPIVAN/loveCodeapp/qryLoveCode",
        "invserv": "/PB2CAPIVAN/invServ/InvServ",
        "donate": "/PB2CAPIVAN/CarInv/Donate",
        "carrier": "/PB2CAPIVAN/Carrier/Aggregate",
    }
    return urljoin(BASE_URL, PATHS[id])


def sign(data: dict, key: str) -> str:
    """Generate signature"""
    data_pairs = [(name, value) for name, value in data.items() if value]
    data_pairs.sort(key=lambda x: x[0])
    query_string = urlencode(data_pairs, quote_via=lambda k, safe, encoding, errors: k)
    signature = b64encode(
        hmac.new(
            key=key.encode("utf-8"),
            msg=query_string.encode("utf-8"),
            digestmod=hashlib.sha256,
        ).digest()
    ).decode("utf-8")
    return signature


def check_api_error(response: Response) -> dict:
    """Check API error"""
    if not isinstance(response, Response):
        raise TypeError("response must be a Response object")
    response.raise_for_status()
    data = response.json()
    if int(data["code"]) != 200:
        raise APIError(data["code"], data["msg"])
    return data


def validate_invoice_number(invoice_number: str) -> bool:
    """Validate einvoice number"""
    if not isinstance(invoice_number, str):
        return False
    return bool(re.match(r"^[A-Z]{2}\d{8}$", invoice_number))


def validate_invoice_random(invoice_random: str) -> bool:
    """Validate invoice random"""
    if not isinstance(invoice_random, str):
        return False
    return bool(re.match(r"^\d{4}$", invoice_random))


def validate_invoice_term(invoice_term: str) -> bool:
    """Validate invoice term"""
    if not isinstance(invoice_term, str):
        return False
    return bool(re.match(r"^\d{3}(02|04|06|08|10|12)$", invoice_term))


def validate_phone_barcode(phone_barcode: str) -> bool:
    """Validate phone barcode"""
    if not isinstance(phone_barcode, str):
        return False
    return bool(re.match(r"^\/[A-Z0-9+.-]{7}$", phone_barcode))



from datetime import date

import pytest

from tw_invoice import AppAPIClient
from tw_invoice.utils import build_api_url

TEST_API_KEY = "test_api_key"
TEST_APP_ID = "test_app_id"
TEST_CARD_ENCRYPT = "3f56c1f14f83b6eb"
TEST_CARD_NUMBER = "/AB12+-."
TEST_CARD_TYPE = "3J0002"
TEST_DATE = date(2020, 1, 1)
TEST_DATE_STR = "2020/01/01"
TEST_INVOICE_ENCRYPT = "a622ef9b2099f91c86dd135a"
TEST_INVOICE_NUMBER = "AB12345678"
TEST_INVOICE_RANDOM = "0123"
TEST_INVOICE_TERM = "10902"
TEST_LOVE_CODE = "0"
TEST_SELLER_ID = "12345678"
TEST_TIME = 1655654400
TEST_TS_TOLERANCE = 30
TEST_UUID = "test_uuid"
TEST_TIMEOUT = (3, 1)


@pytest.fixture
def client():
    return AppAPIClient(TEST_APP_ID, TEST_API_KEY, TEST_UUID)


def test_plain_init():
    client = AppAPIClient(TEST_APP_ID, TEST_API_KEY)
    assert client.app_id == TEST_APP_ID
    assert client.api_key == TEST_API_KEY


# def test_init_with_env_vars():
#     client = AppAPIClient()
#     assert client.app_id == TEST_APP_ID
#     assert client.api_key == TEST_API_KEY


def test_init_with_optional_params():
    client = AppAPIClient(
        app_id=TEST_APP_ID,
        api_key=TEST_API_KEY,
        uuid=TEST_UUID,
        ts_tolerance=TEST_TS_TOLERANCE,
    )
    assert client.app_id == TEST_APP_ID
    assert client.api_key == TEST_API_KEY
    assert client.uuid == TEST_UUID
    assert client.ts_tolerance == TEST_TS_TOLERANCE


def test_init_with_invalid_ts_tolerance():
    with pytest.raises(ValueError):
        AppAPIClient(TEST_APP_ID, TEST_API_KEY, ts_tolerance=0)
    with pytest.raises(ValueError):
        AppAPIClient(TEST_APP_ID, TEST_API_KEY, ts_tolerance=181)


def test_init_with_invalid_skip_validation():
    with pytest.raises(ValueError):
        AppAPIClient(TEST_APP_ID, TEST_API_KEY, skip_validation="invalid")


def test_get_lottery_numbers(client, mocker):
    with pytest.raises(ValueError):
        client.get_lottery_numbers("2022-04-23")

    # Mock the API response
    mocked_session_post = mocker.patch("tw_invoice.app_client.Session.post")
    mocked_check_api_error = mocker.patch("tw_invoice.app_client.check_api_error")
    mocked_parse_obj = mocker.patch(
        "tw_invoice.app_client.LotteryNumberResponse.parse_obj"
    )

    client.get_lottery_numbers("11006")
    mocked_session_post.assert_called_once_with(
        build_api_url("invapp"),
        data={
            "version": 0.2,
            "action": "QryWinningList",
            "invTerm": "11006",
            "UUID": TEST_UUID,
            "appID": TEST_APP_ID,
        },
        timeout=TEST_TIMEOUT,
    )
    mocked_check_api_error.assert_called_once()
    mocked_parse_obj.assert_called_once()


def test_get_invoice_header(client, mocker):
    with pytest.raises(ValueError):
        client.get_invoice_header(
            barcode_type="invalid barcode",
            invoice_number=TEST_INVOICE_NUMBER,
            invoice_date=TEST_DATE,
        )

    with pytest.raises(ValueError):
        client.get_invoice_header(
            barcode_type="QRCode",
            invoice_number="invalid invoice number",
            invoice_date=TEST_DATE,
        )

    for barcode_type in ("QRCode", "Barcode"):
        # Mock the API response
        mocked_validate_invoice_number = mocker.patch(
            "tw_invoice.app_client.validate_invoice_number"
        )
        mocked_session_post = mocker.patch("tw_invoice.app_client.Session.post")
        mocked_check_api_error = mocker.patch("tw_invoice.app_client.check_api_error")
        mocked_parse_obj = mocker.patch(
            "tw_invoice.app_client.InvoiceHeaderResponse.parse_obj"
        )

        client.get_invoice_header(
            barcode_type=barcode_type,
            invoice_number=TEST_INVOICE_NUMBER,
            invoice_date=TEST_DATE,
        )
        mocked_validate_invoice_number.assert_called_once_with(TEST_INVOICE_NUMBER)
        mocked_session_post.assert_called_once_with(
            build_api_url("invapp"),
            data={
                "version": 0.5,
                "type": barcode_type,
                "invNum": TEST_INVOICE_NUMBER,
                "action": "qryInvHeader",
                "generation": "V2",
                "invDate": TEST_DATE_STR,
                "UUID": TEST_UUID,
                "appID": TEST_APP_ID,
            },
            timeout=TEST_TIMEOUT,
        )
        mocked_check_api_error.assert_called_once()
        mocked_parse_obj.assert_called_once()


def test_get_invoice_detail(client, mocker):
    # Test invalid barcode type
    with pytest.raises(ValueError):
        client.get_invoice_detail(
            barcode_type="invalid barcode",
            invoice_number=TEST_INVOICE_NUMBER,
            invoice_date=TEST_DATE,
            invoice_random=TEST_INVOICE_RANDOM,
        )

    # Test QRCode without invoice encrypt
    with pytest.raises(ValueError):
        client.get_invoice_detail(
            barcode_type="QRCode",
            invoice_number=TEST_INVOICE_NUMBER,
            invoice_date=TEST_DATE,
            invoice_random=TEST_INVOICE_RANDOM,
        )

    # Test QRCode invalid invoice encrypt
    with pytest.raises(ValueError):
        client.get_invoice_detail(
            barcode_type="QRCode",
            invoice_number=TEST_INVOICE_NUMBER,
            invoice_date=TEST_DATE,
            invoice_random=TEST_INVOICE_RANDOM,
            invoice_encrypt="invalid invoice encrypt",
        )

    # Test QRCode without seller id
    with pytest.raises(ValueError):
        client.get_invoice_detail(
            barcode_type="QRCode",
            invoice_number=TEST_INVOICE_NUMBER,
            invoice_date=TEST_DATE,
            invoice_random=TEST_INVOICE_RANDOM,
            invoice_encrypt=TEST_INVOICE_ENCRYPT,
        )

    # Test Barcode without invoice term
    with pytest.raises(ValueError):
        client.get_invoice_detail(
            barcode_type="Barcode",
            invoice_number=TEST_INVOICE_NUMBER,
            invoice_date=TEST_DATE,
            invoice_random=TEST_INVOICE_RANDOM,
        )

    # Test invalid invoice term
    with pytest.raises(ValueError):
        client.get_invoice_detail(
            barcode_type="Barcode",
            invoice_number=TEST_INVOICE_NUMBER,
            invoice_date=TEST_DATE,
            invoice_random=TEST_INVOICE_RANDOM,
            invoice_term="invalid invoice term",
        )

    # Test invalid invoice number
    with pytest.raises(ValueError):
        client.get_invoice_detail(
            barcode_type="QRCode",
            invoice_number="invalid invoice number",
            invoice_date=TEST_DATE,
            invoice_random=TEST_INVOICE_RANDOM,
            invoice_encrypt=TEST_INVOICE_ENCRYPT,
            seller_id=TEST_SELLER_ID,
        )

    # Test invalid invoice random
    with pytest.raises(ValueError):
        client.get_invoice_detail(
            barcode_type="QRCode",
            invoice_number=TEST_INVOICE_NUMBER,
            invoice_date=TEST_DATE,
            invoice_random="invalid invoice random",
            invoice_encrypt=TEST_INVOICE_ENCRYPT,
            seller_id=TEST_SELLER_ID,
        )

    # Mock the API response
    mocked_session_post = mocker.patch("tw_invoice.app_client.Session.post")
    mocked_check_api_error = mocker.patch("tw_invoice.app_client.check_api_error")
    mocked_parse_obj = mocker.patch(
        "tw_invoice.app_client.InvoiceDetailResponse.parse_obj"
    )

    client.get_invoice_detail(
        barcode_type="QRCode",
        invoice_number=TEST_INVOICE_NUMBER,
        invoice_date=TEST_DATE,
        invoice_random=TEST_INVOICE_RANDOM,
        invoice_encrypt=TEST_INVOICE_ENCRYPT,
        seller_id=TEST_SELLER_ID,
    )
    mocked_session_post.assert_called_once_with(
        build_api_url("invapp"),
        data={
            "version": 0.6,
            "type": "QRCode",
            "invNum": TEST_INVOICE_NUMBER,
            "action": "qryInvDetail",
            "generation": "V2",
            "invTerm": None,
            "invDate": TEST_DATE_STR,
            "encrypt": TEST_INVOICE_ENCRYPT,
            "sellerID": TEST_SELLER_ID,
            "UUID": TEST_UUID,
            "randomNumber": TEST_INVOICE_RANDOM,
            "appID": TEST_APP_ID,
        },
        timeout=TEST_TIMEOUT,
    )
    mocked_check_api_error.assert_called_once()
    mocked_parse_obj.assert_called_once()

    # Mock the API response
    mocked_session_post = mocker.patch("tw_invoice.app_client.Session.post")
    mocked_check_api_error = mocker.patch("tw_invoice.app_client.check_api_error")
    mocked_parse_obj = mocker.patch(
        "tw_invoice.app_client.InvoiceDetailResponse.parse_obj"
    )

    client.get_invoice_detail(
        barcode_type="Barcode",
        invoice_number=TEST_INVOICE_NUMBER,
        invoice_date=TEST_DATE,
        invoice_random=TEST_INVOICE_RANDOM,
        invoice_term=TEST_INVOICE_TERM,
    )
    mocked_session_post.assert_called_once_with(
        build_api_url("invapp"),
        data={
            "version": 0.6,
            "type": "Barcode",
            "invNum": TEST_INVOICE_NUMBER,
            "action": "qryInvDetail",
            "generation": "V2",
            "invTerm": TEST_INVOICE_TERM,
            "invDate": TEST_DATE_STR,
            "encrypt": None,
            "sellerID": None,
            "UUID": TEST_UUID,
            "randomNumber": TEST_INVOICE_RANDOM,
            "appID": TEST_APP_ID,
        },
        timeout=TEST_TIMEOUT,
    )
    mocked_check_api_error.assert_called_once()
    mocked_parse_obj.assert_called_once()


def test_get_love_code(client, mocker):
    # Mock the API response
    mocked_session_post = mocker.patch("tw_invoice.app_client.Session.post")
    mocked_check_api_error = mocker.patch("tw_invoice.app_client.check_api_error")
    mocked_parse_obj = mocker.patch("tw_invoice.app_client.LoveCodeResponse.parse_obj")

    client.get_love_code("test-query")
    mocked_session_post.assert_called_once_with(
        build_api_url("lovecode"),
        data={
            "version": 0.2,
            "qKey": "test-query",
            "action": "qryLoveCode",
            "UUID": TEST_UUID,
            "appID": TEST_APP_ID,
        },
        timeout=TEST_TIMEOUT,
    )
    mocked_check_api_error.assert_called_once()
    mocked_parse_obj.assert_called_once()


def test_get_carrier_invoices_header(client, mocker):
    # Mock the API response
    mocked_time = mocker.patch("tw_invoice.app_client.time", return_value=TEST_TIME)
    mocked_session_post = mocker.patch("tw_invoice.app_client.Session.post")
    mocked_check_api_error = mocker.patch("tw_invoice.app_client.check_api_error")
    mocked_parse_obj = mocker.patch(
        "tw_invoice.app_client.CarrierInvoicesHeaderResponse.parse_obj"
    )

    client.get_carrier_invoices_header(
        card_type=TEST_CARD_TYPE,
        card_number=TEST_CARD_NUMBER,
        start_date=TEST_DATE,
        end_date=TEST_DATE,
        card_encrypt=TEST_CARD_ENCRYPT,
    )
    mocked_time.assert_called_once()
    mocked_session_post.assert_called_once_with(
        build_api_url("invserv"),
        data={
            "version": 0.5,
            "cardType": TEST_CARD_TYPE,
            "cardNo": TEST_CARD_NUMBER,
            "expTimeStamp": "2147483647",
            "action": "carrierInvChk",
            "timeStamp": TEST_TIME + client.ts_tolerance,
            "startDate": TEST_DATE_STR,
            "endDate": TEST_DATE_STR,
            "onlyWinningInv": "N",
            "uuid": TEST_UUID,
            "appID": TEST_APP_ID,
            "cardEncrypt": TEST_CARD_ENCRYPT,
        },
        timeout=TEST_TIMEOUT,
    )
    mocked_check_api_error.assert_called_once()
    mocked_parse_obj.assert_called_once()


def test_get_carrier_invoices_detail(client, mocker):
    # Test invalid invoice number
    with pytest.raises(ValueError):
        client.get_carrier_invoices_detail(
            card_type=TEST_CARD_TYPE,
            card_number=TEST_CARD_NUMBER,
            invoice_number="invalid invoice number",
            invoice_date=TEST_DATE,
            card_encrypt=TEST_CARD_ENCRYPT,
        )

    # Mock the API response
    mocked_time = mocker.patch("tw_invoice.app_client.time", return_value=TEST_TIME)
    mocked_session_post = mocker.patch("tw_invoice.app_client.Session.post")
    mocked_check_api_error = mocker.patch("tw_invoice.app_client.check_api_error")
    mocked_parse_obj = mocker.patch(
        "tw_invoice.app_client.CarrierInvoicesDetailResponse.parse_obj"
    )

    client.get_carrier_invoices_detail(
        card_type=TEST_CARD_TYPE,
        card_number=TEST_CARD_NUMBER,
        invoice_number=TEST_INVOICE_NUMBER,
        invoice_date=TEST_DATE,
        card_encrypt=TEST_CARD_ENCRYPT,
    )
    mocked_time.assert_called_once()
    mocked_session_post.assert_called_once_with(
        build_api_url("invserv"),
        data={
            "version": 0.5,
            "cardType": TEST_CARD_TYPE,
            "cardNo": TEST_CARD_NUMBER,
            "expTimeStamp": "2147483647",
            "action": "carrierInvDetail",
            "timeStamp": TEST_TIME + client.ts_tolerance,
            "invNum": TEST_INVOICE_NUMBER,
            "invDate": TEST_DATE_STR,
            "uuid": TEST_UUID,
            "sellerName": None,
            "amount": None,
            "appID": TEST_APP_ID,
            "cardEncrypt": TEST_CARD_ENCRYPT,
        },
        timeout=TEST_TIMEOUT,
    )
    mocked_check_api_error.assert_called_once()
    mocked_parse_obj.assert_called_once()


def test_carrier_donate_invoice(client, mocker):
    with pytest.raises(ValueError):
        client.carrier_donate_invoice(
            card_type=TEST_CARD_TYPE,
            card_number=TEST_CARD_NUMBER,
            invoice_date=TEST_DATE,
            invoice_number="invalid invoice number",
            love_code=TEST_LOVE_CODE,
            card_encrypt=TEST_CARD_ENCRYPT,
        )

    # Mock the API response
    mocked_time = mocker.patch("tw_invoice.app_client.time", return_value=TEST_TIME)
    mocked_session_post = mocker.patch("tw_invoice.app_client.Session.post")
    mocked_check_api_error = mocker.patch("tw_invoice.app_client.check_api_error")
    mocked_parse_obj = mocker.patch(
        "tw_invoice.app_client.CarrierInvoiceDonateResponse.parse_obj"
    )

    client.carrier_donate_invoice(
        card_type=TEST_CARD_TYPE,
        card_number=TEST_CARD_NUMBER,
        invoice_date=TEST_DATE,
        invoice_number=TEST_INVOICE_NUMBER,
        love_code=TEST_LOVE_CODE,
        card_encrypt=TEST_CARD_ENCRYPT,
    )
    mocked_time.assert_called_once()
    mocked_session_post.assert_called_once_with(
        build_api_url("donate"),
        data={
            "version": 0.1,
            "serial": "0000000001",
            "cardType": TEST_CARD_TYPE,
            "cardNo": TEST_CARD_NUMBER,
            "expTimeStamp": "2147483647",
            "action": "carrierInvDnt",
            "timeStamp": TEST_TIME + client.ts_tolerance,
            "invDate": TEST_DATE_STR,
            "invNum": TEST_INVOICE_NUMBER,
            "npoBan": TEST_LOVE_CODE,
            "uuid": TEST_UUID,
            "appID": TEST_APP_ID,
            "cardEncrypt": TEST_CARD_ENCRYPT,
            "signature": "Ydmt4alIcCLYG/VBcI+QUdyfgO5Yp8bfX6TPy5xgsVs=",
        },
        timeout=TEST_TIMEOUT,
    )
    mocked_check_api_error.assert_called_once()
    mocked_parse_obj.assert_called_once()
    assert client.serial == 2


def test_get_aggregate_carrier(client, mocker):
    # Mock the API response
    mocked_time = mocker.patch("tw_invoice.app_client.time", return_value=TEST_TIME)
    mocked_session_post = mocker.patch("tw_invoice.app_client.Session.post")
    mocked_check_api_error = mocker.patch("tw_invoice.app_client.check_api_error")
    mocked_parse_obj = mocker.patch(
        "tw_invoice.app_client.AggregateCarrierResponse.parse_obj"
    )

    client.get_aggregate_carrier(
        card_type=TEST_CARD_TYPE,
        card_number=TEST_CARD_NUMBER,
        card_encrypt=TEST_CARD_ENCRYPT,
    )
    mocked_time.assert_called_once()
    mocked_session_post.assert_called_once_with(
        build_api_url("carrier"),
        data={
            "version": 1.0,
            "serial": "0000000001",
            "action": "qryCarrierAgg",
            "cardType": TEST_CARD_TYPE,
            "cardNo": TEST_CARD_NUMBER,
            "cardEncrypt": TEST_CARD_ENCRYPT,
            "appID": TEST_APP_ID,
            "timeStamp": TEST_TIME + client.ts_tolerance,
            "uuid": TEST_UUID,
            "signature": "2pWN1GfP6S7oncE56OJetvpxGlE1tFYHeqgNwHmIuw4=",
        },
        timeout=TEST_TIMEOUT,
    )
    mocked_check_api_error.assert_called_once()
    mocked_parse_obj.assert_called_once()
    assert client.serial == 2



from typing import List, Union

from pydantic import BaseModel

# LotteryNumberResponse


class LotteryNumberResponse(BaseModel):
    v: str
    code: str
    msg: str
    invoYm: str
    superPrizeNo: str
    spcPrizeNo: str
    spcPrizeNo2: Union[str, None] = None
    spcPrizeNo3: Union[str, None] = None
    firstPrizeNo1: str
    firstPrizeNo2: str
    firstPrizeNo3: str
    firstPrizeNo4: Union[str, None] = None
    firstPrizeNo5: Union[str, None] = None
    firstPrizeNo6: Union[str, None] = None
    firstPrizeNo7: Union[str, None] = None
    firstPrizeNo8: Union[str, None] = None
    firstPrizeNo9: Union[str, None] = None
    firstPrizeNo10: Union[str, None] = None
    sixthPrizeNo1: Union[str, None] = None
    sixthPrizeNo2: Union[str, None] = None
    sixthPrizeNo3: Union[str, None] = None
    superPrizeAmt: str
    spcPrizeAmt: str
    firstPrizeAmt: str
    secondPrizeAmt: str
    thirdPrizeAmt: str
    fourthPrizeAmt: str
    fifthPrizeAmt: str
    sixthPrizeAmt: str
    sixthPrizeNo4: Union[str, None] = None
    sixthPrizeNo5: Union[str, None] = None
    sixthPrizeNo6: Union[str, None] = None


# InvoiceHeaderResponse


class InvoiceHeaderResponse(BaseModel):
    v: str
    code: str
    msg: str
    invNum: str
    invDate: str
    sellerName: str
    invStatus: str
    invPeriod: str
    sellerBan: str
    sellerAddress: Union[str, None] = None
    invoiceTime: str
    buyerBan: Union[str, None] = None
    currency: Union[str, None] = None


# InvoiceDetailResponse


class InvoiceDetail(BaseModel):
    rowNum: str
    description: str
    quantity: str
    unitPrice: str
    amount: str


class InvoiceDetailResponse(BaseModel):
    # v: str  # For some reason this is not in the response
    code: str
    msg: str
    invNum: str
    invDate: str
    sellerName: str
    invStatus: str
    invPeriod: str
    sellerBan: str
    sellerAddress: str
    invoiceTime: str
    buyerBan: str
    currency: str
    amount: str
    details: Union[List[InvoiceDetail], None] = None


# LoveCodeResponse


class LoveCode(BaseModel):
    rowNum: int
    SocialWelfareBAN: str
    LoveCode: str
    SocialWelfareName: str
    SocialWelfareAbbrev: Union[str, None] = None


class LoveCodeResponse(BaseModel):
    v: str
    code: str
    msg: str
    details: List[LoveCode]


# CarrierInvoicesHeaderResponse


class InvoiceDate(BaseModel):
    year: int
    month: int
    date: int
    day: int
    hours: int
    minutes: int
    seconds: int
    time: int
    timezoneOffset: int


class Invoice(BaseModel):
    rowNum: str
    invNum: str
    cardType: str
    cardNo: str
    sellerName: str
    invStatus: str
    invDonatable: bool
    amount: str
    invPeriod: str
    donateMark: bool  # served in 0 or 1, will be cast implicitly to bool
    sellerBan: str
    sellerAddress: Union[str, None] = None
    invoiceTime: str
    buyerBan: Union[str, None] = None
    currency: Union[str, None] = None
    invDate: InvoiceDate


class CarrierInvoicesHeaderResponse(BaseModel):
    v: str
    code: int
    msg: str
    onlyWinningInv: str
    details: List[Invoice]


# CarrierInvoicesDetailResponse


class CarrierInvoicesDetailResponse(BaseModel):
    v: str
    code: int
    msg: str
    invNum: str
    invDate: str
    sellerName: str
    amount: str
    invStatus: str
    invPeriod: str
    details: List[InvoiceDetail]
    sellerBan: str
    sellerAddress: str
    invoiceTime: str
    currency: str


# CarrierInvoiceDonateResponse


class CarrierInvoiceDonateResponse(BaseModel):
    v: str
    code: int
    msg: str
    hashSerial: str
    invNum: str
    invDate: str
    NPOBan: str
    invStatus: str
    invDntTimeStamp: str


# AggregateCarrierResponse


class Carrier(BaseModel):
    carrierType: str
    carrierId2: str
    carrierName: str


class AggregateCarrierResponse(BaseModel):
    v: str
    code: int
    hashSerial: str
    msg: str
    cardType: str
    cardNo: str
    carriers: List[Carrier]
